import numpy as np
import pandas as pd
import starfile
from cryotypes.poseset import PoseSetDataLabels as PSDL
from cryotypes.poseset import validate_poseset_dataframe
from scipy.spatial.transform import Rotation

from ..utils.constants import Relion
from ..utils.generic import ParseError, guess_name_vec


def construct_poseset(
    df,
    coord_headers,
    shift_headers,
    pixel_size_headers,
    exp_header,
    star_path,
    rescale_shifts=False,
    name_regex=None,
    guess_id=True,
):
    coords = np.asarray(df.get(coord_headers, 0), dtype=float)
    ndim = len(coord_headers)
    shifts = np.asarray(df.get(shift_headers, 0), dtype=float)

    pixel_size = df.get(pixel_size_headers, None)
    if pixel_size is not None:
        pixel_size = np.asarray(pixel_size, dtype=float)
        # XXX TODO: remove the following and support variable pixel sizes?
        pixel_size = pixel_size.ravel()[0]
        if rescale_shifts:
            shifts = shifts / pixel_size
    elif rescale_shifts:
        raise ParseError(
            "pixel size information is missing, but it is required because shifts are in Angstroms!"
        )

    if Relion.EULER_HEADERS[0] in df.columns:
        euler_convention = Relion.EULER
        eulers = np.asarray(df.get(Relion.EULER_HEADERS, 0), dtype=float)
    else:
        euler_convention = Relion.INPLANE
        eulers = np.asarray(df.get(Relion.EULER_HEADERS[2], 0), dtype=float)

    rot = Rotation.from_euler(euler_convention, eulers, degrees=True)

    # we want the inverse, which when applied to basis vectors it gives us the particle orientation
    rot = rot.inv()

    features = df.drop(columns=Relion.REDUNDANT_HEADERS, errors="ignore")

    exp_id = df[exp_header]
    if guess_id:
        exp_id = guess_name_vec(exp_id, name_regex)

    data = pd.DataFrame()
    data[PSDL.POSITION[:ndim]] = coords
    data[PSDL.SHIFT[:ndim]] = -shifts  # relion subtracts from coords
    data[PSDL.ORIENTATION] = np.asarray(rot)
    data[PSDL.PIXEL_SPACING] = pixel_size or 0
    data[PSDL.EXPERIMENT_ID] = exp_id
    data[PSDL.SOURCE] = star_path
    data = pd.concat([data, features], axis=1)

    return validate_poseset_dataframe(data, coerce=True)


def parse_relion_star(
    df,
    star_path="",
    name_regex=None,
    guess_id=True,
    **kwargs,
):
    # find out which columns are present and which version we're dealing with
    rescale_shifts = True

    if Relion.PIXEL_SIZE_HEADER["3.1"] in df.columns:
        # 3.1 has priority over 4 (cause it's still used in 4 for single particle)
        version = "3.1"
    elif Relion.PIXEL_SIZE_HEADER["4.0"] in df.columns:
        version = "4.0"
    else:
        # also includes if this is not present (required for later versions!)
        version = "3.0"
        rescale_shifts = False

    pixel_size_headers = Relion.PIXEL_SIZE_HEADER[version]
    shift_headers = Relion.SHIFT_HEADERS[version]

    if (
        Relion.COORD_HEADERS[0] not in df.columns
        or Relion.COORD_HEADERS[1] not in df.columns
    ):
        raise ParseError("coordinate information missing")
    elif Relion.COORD_HEADERS[-1] in df.columns:
        coord_headers = Relion.COORD_HEADERS
    else:
        coord_headers = Relion.COORD_HEADERS[:2]
        shift_headers = shift_headers[:2]

    # start parsing
    if Relion.MICROGRAPH_NAME_HEADER[version] in df.columns:
        exp_header = Relion.MICROGRAPH_NAME_HEADER[version]
    else:
        exp_header = None

    return construct_poseset(
        df,
        coord_headers,
        shift_headers,
        pixel_size_headers,
        exp_header,
        star_path,
        rescale_shifts=rescale_shifts,
        name_regex=name_regex,
        guess_id=guess_id,
    )


def read_star(star_path, **kwargs):
    try:
        raw_data = starfile.read(star_path, always_dict=True)
    except pd.errors.EmptyDataError:  # raised sometimes by .star files with completely different data
        raise ParseError(f"the contents of {star_path} have the wrong format")

    if len(raw_data) == 1:
        df = list(raw_data.values())[0]
    elif "particles" in raw_data and "optics" in raw_data:
        df = raw_data["particles"].merge(
            raw_data["optics"], on=Relion.OPTICS_GROUP_HEADER
        )
    else:
        raise ParseError(
            f"Failed to parse {star_path} as particles. Are you sure this is a particle file?"
        )

    return parse_relion_star(df, star_path=star_path, **kwargs)



import starfile
from cryotypes.poseset import PoseSetDataLabels as PSDL

from cryohub.reading.star import read_star
from cryohub.utils.testing import assert_dataframe_equal

base_columns = [
    *PSDL.POSITION,
    *PSDL.SHIFT,
    PSDL.ORIENTATION,
    PSDL.EXPERIMENT_ID,
    PSDL.PIXEL_SPACING,
    "feature",
]


def test_read_relion30_3d(tmp_path, relion30_star, poseset):
    file_path = tmp_path / "test.star"
    starfile.write(relion30_star, file_path)

    part = read_star(file_path, name_regex=r"\w")

    assert_dataframe_equal(part, poseset, columns=base_columns)


def test_read_relion31_3d(tmp_path, relion31_star, poseset):
    file_path = tmp_path / "test.star"
    starfile.write(relion31_star, file_path)

    part = read_star(file_path, name_regex=r"\w")

    assert_dataframe_equal(part, poseset, columns=base_columns)


def test_read_relion40_3d(tmp_path, relion40_star, poseset):
    file_path = tmp_path / "test.star"
    starfile.write(relion40_star, file_path)

    part = read_star(file_path, name_regex=r"\w")

    assert_dataframe_equal(part, poseset, columns=base_columns)


def test_read_relion40_2d(tmp_path, relion40_star2D, poseset2D):
    file_path = tmp_path / "test.star"
    starfile.write(relion40_star2D, file_path)

    part = read_star(file_path, name_regex=r"\w")

    assert_dataframe_equal(
        part,
        poseset2D,
        columns=[
            *PSDL.POSITION[:2],
            *PSDL.SHIFT[:2],
            PSDL.ORIENTATION,
            PSDL.EXPERIMENT_ID,
            PSDL.PIXEL_SPACING,
        ],
    )



import numpy as np
import pandas as pd
from cryotypes.poseset import PoseSetDataLabels as PSDL
from cryotypes.poseset import validate_poseset_dataframe
from scipy.spatial.transform import Rotation

from ..utils.constants import Relion
from ..utils.generic import guess_name


def read_box(
    path,
    name_regex=None,
    **kwargs,
):
    coords = np.loadtxt(path)
    rot = Rotation.identity(len(coords)).as_euler(Relion.EULER)
    name = guess_name(path, name_regex)

    df = pd.DataFrame(
        {
            PSDL.POSITION: coords,
            PSDL.ORIENTATION: rot,
            PSDL.EXPERIMENT_ID: name,
        }
    )
    return validate_poseset_dataframe(df, coerce=True)



import pandas as pd
import starfile

from cryohub.utils.constants import Relion
from cryohub.utils.testing import assert_dataframe_equal
from cryohub.writing.star import write_star

version = "4.0"

base_columns = [
    *Relion.COORD_HEADERS,
    *Relion.SHIFT_HEADERS[version],
    *Relion.EULER_HEADERS,
    Relion.PIXEL_SIZE_HEADER[version],
    Relion.MICROGRAPH_NAME_HEADER[version],
]


def test_write_star3D(tmp_path, poseset, relion40_star):
    file_path = tmp_path / "test.star"

    write_star(poseset, file_path)
    data = starfile.read(file_path)
    expected = pd.merge(
        relion40_star["particles"], relion40_star["optics"], on="rlnOpticsGroup"
    )
    assert_dataframe_equal(data, expected, columns=base_columns + ["feature"])


def test_write_star2D(tmp_path, poseset2D, relion40_star2D):
    file_path = tmp_path / "test.star"

    write_star(poseset2D, file_path)
    data = starfile.read(file_path)
    expected = pd.merge(
        relion40_star2D["particles"], relion40_star2D["optics"], on="rlnOpticsGroup"
    )
    assert_dataframe_equal(data, expected, columns=base_columns)



import re
from pathlib import Path
from typing import Iterable

import numpy as np
from cryotypes.image import ImageProtocol
from cryotypes.poseset import PoseSet


class ParseError(RuntimeError):
    pass


# a list of commonly used base names for paths in regex form
common_name_regexes = (
    r"\w+_\d+",
    r"\d+",
)


def guess_name(string, name_regex=None):
    """
    guess an appropriate name based on the input string
    and a list of regexes in order of priority
    """
    if string is None:
        return ""
    if isinstance(string, Path):
        string = string.stem
    name_regex = listify(name_regex)
    regexes = list(common_name_regexes)
    regexes = name_regex + regexes
    for regex in regexes:
        if match := re.search(regex, str(string)):
            if match.groups():
                return match.group(1)
            return match.group(0)
    else:
        return ""


@np.vectorize
def guess_name_vec(string, reg=None):
    return guess_name(string, reg)


def pad_to_3D(arr):
    if arr.shape[-1] == 2:
        arr = np.pad(arr, ((0, 0), (0, 1)))
    return arr


def listify(obj):
    """
    transform input into an appropriate list, unless already list-like
    """
    if isinstance(obj, Iterable):
        if isinstance(obj, (str, Path, PoseSet, ImageProtocol)):
            return [obj]
        else:
            return list(obj)
    if obj is None:
        return []
    return [obj]



import dynamotable
import numpy as np
import pandas as pd
from cryotypes.poseset import PoseSetDataLabels as PSDL
from cryotypes.poseset import validate_poseset_dataframe
from scipy.spatial.transform import Rotation

from ..utils.constants import Dynamo
from ..utils.generic import guess_name_vec


def read_tbl(table_path, table_map_file=None, guess_id=True, name_regex=None, **kwargs):
    """
    Read particles from a dynamo format table file
    """
    df = dynamotable.read(table_path, table_map_file)

    if Dynamo.EXP_NAME_HEADER in df.columns:
        exp_id = df[Dynamo.EXP_NAME_HEADER]
        if guess_id:
            exp_id = guess_name_vec(exp_id, name_regex)
    else:
        exp_id = df[Dynamo.EXP_ID_HEADER].astype(str)

    if Dynamo.COORD_HEADERS[-1] in df.columns:
        ndim = 3
    else:
        ndim = 2

    coords = np.asarray(df[Dynamo.COORD_HEADERS[:ndim]], dtype=float)
    shifts = np.asarray(df.get(Dynamo.SHIFT_HEADERS[:ndim], 0), dtype=float)
    eulers = np.asarray(df.get(Dynamo.EULER_HEADERS[ndim], 0), dtype=float)

    if ndim == 3:
        rot = Rotation.from_euler(Dynamo.EULER, eulers, degrees=True)
    else:
        rot = Rotation.from_euler(Dynamo.INPLANE, eulers, degrees=True)

    # we want the inverse, which when applied to basis vectors it gives us the particle orientation
    rot = rot.inv()

    features = df.drop(columns=Dynamo.REDUNDANT_HEADERS, errors="ignore")

    data = pd.DataFrame()
    data[PSDL.POSITION[:ndim]] = coords
    data[PSDL.SHIFT[:ndim]] = shifts
    data[PSDL.ORIENTATION] = np.asarray(rot)
    data[PSDL.EXPERIMENT_ID] = exp_id
    data[PSDL.SOURCE] = table_path
    data = pd.concat([data, features], axis=1)

    return validate_poseset_dataframe(data, coerce=True)



import pandas as pd
import starfile
from cryotypes.poseset import PoseSetDataLabels as PSDL
from cryotypes.poseset import validate_poseset_dataframe
from scipy.spatial.transform import Rotation

from ..utils.constants import Relion
from ..utils.generic import guess_name


def read_cbox(
    cbox_path,
    name_regex=None,
    **kwargs,
):
    data = starfile.read(cbox_path)["cryolo"]
    coords = data[[f"Coordinate{axis}" for axis in "XYZ"]].to_numpy()
    rot = Rotation.identity(len(coords)).as_euler(Relion.EULER)
    name = guess_name(cbox_path, name_regex)

    df = pd.DataFrame(
        {
            PSDL.POSITION: coords,
            PSDL.ORIENTATION: rot,
            PSDL.EXPERIMENT_ID: name,
        }
    )
    return validate_poseset_dataframe(df, coerce=True)



import dynamotable
import numpy as np
import pandas as pd
from cryotypes.poseset import PoseSetDataLabels as PSDL
from scipy.spatial.transform import Rotation

from ..utils.constants import Dynamo


def write_tbl(particles, file_path):
    """
    write particle data to disk as a dynamo .tbl file
    """
    ndim = 3 if PSDL.POSITION_Z in particles.columns else 2

    df = pd.DataFrame()
    df[Dynamo.COORD_HEADERS[:ndim]] = particles[PSDL.POSITION[:ndim]]
    df[Dynamo.SHIFT_HEADERS[:ndim]] = particles[PSDL.SHIFT[:ndim]]

    rot = Rotation.concatenate(particles[PSDL.ORIENTATION]).inv()  # invert back
    eulers = rot.as_euler(Dynamo.EULER, degrees=True)
    if np.allclose(eulers[:, 1:], 0):
        # single angle world
        df[Dynamo.EULER_HEADERS[2]] = eulers[:, 0]
    else:
        df[Dynamo.EULER_HEADERS[3]] = eulers

    df[Dynamo.EXP_ID_HEADER] = particles[PSDL.EXPERIMENT_ID]

    if not str(file_path).endswith(".tbl"):
        file_path = str(file_path) + ".tbl"
    dynamotable.write(df, file_path)



import dynamotable
from cryotypes.poseset import PoseSetDataLabels as PSDL

from cryohub.reading.tbl import read_tbl
from cryohub.utils.testing import assert_dataframe_equal

base_columns = [
    *PSDL.POSITION,
    *PSDL.SHIFT,
    PSDL.ORIENTATION,
]


def test_read_tbl(tmp_path, dynamo_tbl, poseset):
    file_path = tmp_path / "test.tbl"
    dynamotable.write(dynamo_tbl, file_path)

    part = read_tbl(file_path, name_regex=r"\w")

    assert_dataframe_equal(part, poseset, columns=base_columns)



import numpy as np
import pandas as pd
import starfile
from cryotypes.poseset import PoseSetDataLabels as PSDL
from scipy.spatial.transform import Rotation

from ..utils.constants import POSESET_REDUNDANT_HEADERS, Relion


def write_star(particles, file_path, version="4.0", overwrite=False):
    """
    write particle data to disk as a .star file
    """
    ndim = 3 if PSDL.POSITION_Z in particles.columns else 2

    df = pd.DataFrame()
    df[Relion.COORD_HEADERS[:ndim]] = particles[PSDL.POSITION[:ndim]]

    px_size = particles[PSDL.PIXEL_SPACING]
    shifts = particles[PSDL.SHIFT[:ndim]]
    if version != "3.0":
        # shifts are in Angstroms (we need to go to numpy and resize or indices mess up stuff)
        shifts *= px_size.to_numpy().reshape(len(px_size), -1)
    df[Relion.PIXEL_SIZE_HEADER[version]] = px_size
    df[
        Relion.SHIFT_HEADERS[version][:ndim]
    ] = -shifts  # shifts are subtractive in relion

    rot = Rotation.concatenate(particles[PSDL.ORIENTATION]).inv()
    eulers = rot.as_euler(Relion.EULER, degrees=True)  # invert for relion
    if np.allclose(eulers[:, 1:], 0):
        # single angle world
        df[Relion.EULER_HEADERS[2]] = eulers[:, 0]
    else:
        df[Relion.EULER_HEADERS] = eulers
    df[Relion.MICROGRAPH_NAME_HEADER[version]] = particles[PSDL.EXPERIMENT_ID]

    features = particles.drop(columns=POSESET_REDUNDANT_HEADERS, errors="ignore")
    df = pd.concat([df, features], axis=1)

    # split out optics group if present
    if Relion.OPTICS_GROUP_HEADER in particles.columns:
        optics_headers = [
            h for h in Relion.POSSIBLE_OPTICS_GROUP_HEADERS if h in particles.columns
        ]
        optics = (
            df.get([Relion.OPTICS_GROUP_HEADER] + optics_headers)
            .drop_duplicates()
            .reset_index(drop=True)
        )
        particles = df.drop(columns=optics_headers, errors="ignore")
        data = {"optics": optics, "particles": particles}
    else:
        data = df

    if not str(file_path).endswith(".star"):
        file_path = str(file_path) + ".star"

    starfile.write(data, file_path, overwrite=overwrite)
